from __future__ import annotations

from typing import Any
from typing import Container
from typing import Iterable
from typing import Mapping
from typing import Sized
from typing import Union
from typing import cast
from typing import get_args
from typing import get_origin

from check_types.errors import ClassArgumentIsAParametrizedContainer
from check_types.errors import ClassArgumentIsNotAClass
from check_types.errors import SpecificTypeError
from check_types.errors import TypeErrorsInIterable
from check_types.errors import TypeErrorsInMapping


def check_isinstance(obj: Any, cls: Any) -> None:
    if origin := get_origin(cls):
        if origin is Union:
            _check_union(obj, cls)
        elif issubclass(origin, Mapping):
            _check_mapping(obj, cls, origin)
        elif issubclass(origin, Iterable):
            _check_iterable(obj, cls, origin)
        elif issubclass(origin, Sized):
            _check_sized(obj, cls, origin)
        elif issubclass(origin, Container):
            _check_container(obj, cls, origin)
        else:
            raise NotImplementedError(f"{origin=}")
    elif not isinstance(cls, type):
        raise ClassArgumentIsNotAClass(cls)
    elif (
        isinstance(obj, bool)
        and issubclass(cls, int)
        and not issubclass(cls, bool)
    ):
        raise SpecificTypeError(obj, cls)
    elif not isinstance(obj, cls):
        raise SpecificTypeError(obj, cls)


def _check_container(obj: Any, cls: Any, origin: type[Container]) -> None:
    if get_args(cls):
        raise ClassArgumentIsAParametrizedContainer(cls)
    elif not isinstance(obj, origin):
        raise SpecificTypeError(obj, cls)


def _check_mapping(obj: Any, cls: Any, origin: type[Mapping]) -> None:
    if not isinstance(obj, origin):
        raise SpecificTypeError(obj, cls)
    if args := get_args(cls):
        key_cls, value_cls = args
        key_errors, value_errors = {}, {}
        for key, value in obj.items():
            try:
                check_isinstance(key, key_cls)
            except TypeError as error:
                key_errors[key] = error
            try:
                check_isinstance(value, value_cls)
            except TypeError as error:
                value_errors[key] = error
        if key_errors or value_errors:
            raise TypeErrorsInMapping(key_errors, value_errors)


def _check_iterable(obj: Any, cls: Any, origin: type[Iterable]) -> None:
    if not isinstance(obj, origin):
        raise SpecificTypeError(obj, cls)
    args = get_args(cls)
    errors = {}
    if issubclass(origin, tuple):
        obj = cast(tuple, obj)
        if len(args) == 2 and args[1] is Ellipsis:
            inner, _ = args
            for i, element in enumerate(obj):
                try:
                    check_isinstance(element, inner)
                except TypeError as error:
                    errors[i] = error
        elif len(obj) == len(args):
            for i, (element, inner) in enumerate(zip(obj, args)):
                try:
                    check_isinstance(element, inner)
                except TypeError as error:
                    errors[i] = error
        else:
            raise SpecificTypeError(obj, cls)
    elif len(args) == 1:
        (inner,) = args
        for i, element in enumerate(obj):
            try:
                check_isinstance(element, inner)
            except TypeError as error:
                errors[i] = error
    if errors:
        raise TypeErrorsInIterable(errors)


def _check_sized(obj: Any, cls: Any, origin: type[Sized]) -> None:
    if not isinstance(obj, origin):
        raise SpecificTypeError(obj, cls)


def _check_union(obj: Any, cls: Any) -> None:
    if not any(isinstance_plus(obj, cls_i) for cls_i in get_args(cls)):
        raise SpecificTypeError(obj, cls)


def isinstance_plus(obj: Any, cls: Any) -> bool:
    try:
        check_isinstance(obj, cls)
    except TypeError:
        return False
    else:
        return True
