from __future__ import annotations

from typing import Any
from typing import Generic
from typing import Mapping
from typing import TypeVar


T, U = TypeVar("T"), TypeVar("U")


class _BaseTypeError(TypeError):
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, type(self)):
            return self._check_equal(other)
        else:
            return NotImplemented

    def _check_equal(self: T, _other: T) -> bool:
        raise NotImplementedError(f"{self=}")


class ArgumentIsAClass(_BaseTypeError):
    def __init__(self, obj: Any) -> None:
        self.obj = obj

    def _check_equal(self, other: ArgumentIsAClass) -> bool:
        return self.obj == other.obj


class ArgumentIsNotADataClass(_BaseTypeError):
    def __init__(self, obj: Any) -> None:
        self.obj = obj

    def _check_equal(self, other: ArgumentIsNotADataClass) -> bool:
        return self.obj == other.obj


class ClassArgumentIsAParametrizedContainer(_BaseTypeError):
    def __init__(self, cls: Any) -> None:
        self.cls = cls

    def _check_equal(
        self, other: ClassArgumentIsAParametrizedContainer
    ) -> bool:
        return self.cls == other.cls


class ClassArgumentIsNotAClass(_BaseTypeError):
    def __init__(self, cls: Any) -> None:
        self.cls = cls

    def _check_equal(self, other: ClassArgumentIsNotAClass) -> bool:
        return self.cls == other.cls


class SpecificTypeError(_BaseTypeError):
    def __init__(self, obj: Any, cls: Any) -> None:
        self.obj = obj
        self.cls = cls

    def _check_equal(self, other: SpecificTypeError) -> bool:
        return self.obj == other.obj and self.cls == other.cls


class TypeErrorsInDataClass(_BaseTypeError):
    def __init__(self, errors: Mapping[str, _BaseTypeError]) -> None:
        self.errors = dict(errors)

    def _check_equal(self, other: TypeErrorsInDataClass) -> bool:
        return self.errors == other.errors


class TypeErrorsInIterable(_BaseTypeError):
    def __init__(self, errors: Mapping[int, _BaseTypeError]) -> None:
        self.errors = dict(errors)

    def _check_equal(self, other: TypeErrorsInIterable) -> bool:
        return self.errors == other.errors


class TypeErrorsInMapping(_BaseTypeError, Generic[T, U]):
    def __init__(
        self,
        keys: Mapping[T, _BaseTypeError],
        values: Mapping[U, _BaseTypeError],
    ) -> None:
        self.keys = dict(keys)
        self.values = dict(values)

    def _check_equal(self, other: TypeErrorsInMapping) -> bool:
        return self.keys == other.keys and self.values == other.values
